-------------------------------------------------------------------------------
-- EmoteSplitter by Tammya-MoonGuard
--
-- Allows you to easily paste long emotes in the chat window.
-------------------------------------------------------------------------------

local Main = LibStub("AceAddon-3.0"):NewAddon( "EmoteSplitter", 
		"AceHook-3.0", "AceTimer-3.0", "AceEvent-3.0" )
		
EmoteSplitter = Main

local g_chat_queue = {}
local g_chat_timer = nil
local g_chat_busy  = false

local g_ctl_hook_sendchat
--local g_ctl_hook_update

local g_inside_ctl

-- fastpost
local FASTPOST_PERIOD = 0.5 -- period between fastposts allowed

Main.messages_waiting = 0
Main.fastpost_time = 0

-------------------------------------------------------------------------------
function Main:OnInitialize()
end 

-------------------------------------------------------------------------------
SlashCmdList["EMOTESPLITTER"] = function( msg ) 
	Main:Options_Show()
end
 
-------------------------------------------------------------------------------
function Main:OnEnable() 

	
	if UCM then
		print( "EmoteSplitter cannot run with UnlimitedChatMessage enabled." )
		return
	end
	
	self:Options_Init()
	SLASH_EMOTESPLITTER1 = "/emotesplitter"
	
	-- Hook events to verify that chat messages are sent.
	self:RegisterEvent( "CHAT_MSG_SAY", "OnChatMsgSay" )
	self:RegisterEvent( "CHAT_MSG_EMOTE", "OnChatMsgEmote" )
	self:RegisterEvent( "CHAT_MSG_YELL", "OnChatMsgYell" )
	self:RegisterEvent( "CHAT_MSG_SYSTEM", "OnChatMsgSystem" ) 
	
	-- Chat hooks for splitting messages.
	self:RawHook( "SendChatMessage", true )
	self:RawHook( "BNSendWhisper", true )
	
	-- Unlock chat boxes.
	for i = 1, NUM_CHAT_WINDOWS do
		_G["ChatFrame" .. i .. "EditBox"]:SetMaxLetters( 0 )
		_G["ChatFrame" .. i .. "EditBox"]:SetMaxBytes( 0 )
	end 
	
	-- Hook ChatThrottleLib
	g_ctl_hook_sendchat = ChatThrottleLib.SendChatMessage
	--g_ctl_hook_update = ChatThrottleLib.OnUpdate
	ChatThrottleLib.SendChatMessage = Main.CTL_SendChatMessage
	--ChatThrottleLib.OnUpdate = Main.CTL_OnUpdate
	ChatThrottleLib.Frame:SetScript( "OnUpdate", Main.CTL_OnUpdate )
end

-------------------------------------------------------------------------------
function Main.CTL_SendChatMessage( ... )
	-- organic call to ChatThrottleLib
	
	local ctl, _, _, text, chattype, language = ...
	chattype = chattype:upper()
	if chattype == "SAY" or chattype == "EMOTE" or chattype == "YELL" then
		self:SendChat( text, chattype, language )
		return
	end
	
	g_inside_ctl = true
	g_ctl_hook_sendchat( ... )
	g_inside_ctl = false
end

-------------------------------------------------------------------------------
function Main.CTL_OnUpdate( ... )
	g_inside_ctl = true
	--g_ctl_hook_update()
	ChatThrottleLib.OnUpdate( ... )
	g_inside_ctl = false
end

-------------------------------------------------------------------------------
-- Hook for Battle.net whispers.
-------------------------------------------------------------------------------
function Main:BNSendWhisper( presenceID, messageText ) 
	
	if messageText:len() > 255 then
		local chunks = self:SplitMessage( messageText )
		for k,v in ipairs( chunks ) do
			self.hooks.BNSendWhisper( presenceID, v )
		end
		return
	end
	
	self.hooks.BNSendWhisper( presenceID, messageText )
end
 
-------------------------------------------------------------------------------
-- Hook for chat message sending.
-------------------------------------------------------------------------------
function Main:SendChatMessage( msg, chatType, language, channel )

	if g_inside_ctl then
		-- we're inside CTL, don't try to edit the message
		self.hooks.SendChatMessage( msg, chatType, language, channel )
		return
	end
	
	-- otherwise, this is an organic call,
	-- and we want to adjust it and transfer it to our queue and CTL
	
	if msg:find( "\n" ) then
		for splitmsg in msg:gmatch( "[^\n]+" ) do
			self:SendChatMessage( splitmsg, chatType, language, channel )
		end
		return
	end
	
	local chunks = self:SplitMessage( msg )

	for i = 1, #chunks do
		self:SendChat( chunks[i], chatType, language, channel )
	end
end

-------------------------------------------------------------------------------
function Main:SplitMessage( text )
	local replaced_links = {}
	local chunks = {}
	
	if text:len() < 255 then
		-- fits in a single message.
		table.insert( chunks, text )
		return chunks
	end
	
	-- replace links with filler
	text = text:gsub( "(|H[^|]+|h[^|]+|h)", function( link )
			table_insert( replaced_links, link );
			return "\001\002" .. string.rep( "p", link:len() - 4 ) .. "\003\004";
			-- note that links without trailing or leading spaces can break things
	end)
	
	-- todo: replace all kinds of spaces with actual spaces.
	
	local premark = self.db.global.premark
	local postmark = self.db.global.postmark
	
	if premark ~= "" then premark = premark .. " " end
	if postmark ~= "" then postmark = " " .. postmark end
	
	while( text:len() > 255 ) do
		-- we could start at 256, but that's scary :)
		
		for i = 255 - postmark:len(), 1, -1 do
			if string.byte( text, i ) == 32 then -- space
				
				table.insert( chunks, text:sub( 1, i-1 ) .. postmark )
				text = premark .. text:sub( i+1 )
				break
			end
			
			if i == 1 then
				-- this may break utf8 characters, but we don't give a shit because it's an abuse case
				table.insert( chunks, text:sub( 1, 255 - postmark:len() ) .. postmark )
				text = premark .. text:sub( 255 - postmark:len() + 1 )
				break
			end
			
		end
	end
	
	-- and the final chunk
	table.insert( chunks, text )
	
	-- put links back
	local link_count = 0
	for i = 1, #chunks do
		chunks[i] = chunks[i]:gsub("\001\002p+\003\004", function(link)
			link_count = link_count + 1
			return replaced_links[link_count] 
		end)
	end
	
	return chunks
end

-------------------------------------------------------------------------------
local function OnCTL_Sent()
	local self = Main
	self.messages_waiting = self.messages_waiting - 1
end
 
-------------------------------------------------------------------------------
function Main:SendChat( msg, type, lang, channel )
	type = type:upper()
	
	
	if type == "SAY" or type == "EMOTE" or type == "YELL" then
		if msg == "" then return end
		if msg:find( "卍" ) or msg:find( "卐" ) then return end -- sending these silently fails
		
		-- say and emote have problematic throttling
		table.insert( g_chat_queue, { msg=msg, type=type, lang=lang } )
		self:StartChat()
	else
		
		self:CommitChat( msg, type, lang, channel )
		--[[
		g_inside_ctl = true
		
		if g_messages_queued == 1 then
			-- fast opst
			self.hooks.SendChatMessage( msg, chatType, language, channel )
		else
		
			g_ctl_hook_sendchat( ChatThrottleLib, "ALERT", "EMS", msg, type, lang, channel, nil, OnCTL_Sent );
		end
		g_inside_ctl = false]]
	end
end

-------------------------------------------------------------------------------
function Main:CommitChat( msg, type, lang, channel )
	self.messages_waiting = self.messages_waiting + 1
	if self.db.global.fastpost and self.messages_waiting == 1 
	   and GetTime() - self.fastpost_time > FASTPOST_PERIOD then
	   
		self.fastpost_time = GetTime()
		self.messages_waiting = 0
		self.hooks.SendChatMessage( msg, type, language, channel )
	else
		g_inside_ctl = true
		g_ctl_hook_sendchat( ChatThrottleLib, "ALERT", "EMS", msg, type, lang, channel, nil, OnCTL_Sent );
		g_inside_ctl = false
	end
end

-------------------------------------------------------------------------------
function Main:StartChat()
	if g_chat_busy then return end
	if #g_chat_queue == 0 then return end
	g_chat_busy = true
	
	self:ChatQueue()
end

-------------------------------------------------------------------------------
function Main:ChatQueue()
	 
	if #g_chat_queue == 0 then 
		g_chat_busy = false
		return 
	end
	
	local c = g_chat_queue[1]
	
	self:CommitChat( c.msg, c.type, c.lang )
	--g_inside_ctl = true
	--g_ctl_hook_sendchat( ChatThrottleLib, "ALERT", "EMS", c.msg, c.type, c.lang );
	--g_inside_ctl = false
	
	g_chat_timer = self:ScheduleTimer( "ChatTimeout", 10 )
end

-------------------------------------------------------------------------------
function Main:ChatTimeout()
	g_chat_timer = nil
	g_chat_queue = {}
	g_chat_busy = false
	print( "|cffff0000<Chat failed!>|r" )
end

-------------------------------------------------------------------------------
function Main:ChatConfirmed()
	self:CancelTimer( g_chat_timer )
	g_chat_timer = nil
	table.remove( g_chat_queue, 1 )
	
	self:ChatQueue()
end

-------------------------------------------------------------------------------
function Main:ChatFailed()
	self:CancelTimer( g_chat_timer )
	print( "|cffff0000<Chat failed; waiting...>" )
	g_chat_timer = self:ScheduleTimer( "ChatFailedRetry", 3 )
end

-------------------------------------------------------------------------------
function Main:ChatFailedRetry()
	print( "|cffff00ff<Resending...>" )
	g_chat_timer = nil
	self:ChatQueue()
end

-------------------------------------------------------------------------------
function Main:OnChatMsgSay( event, message, sender, _, _, _, _, _, _, _, _, _, guid )

	if #g_chat_queue == 0 then return end
	
	local cq = g_chat_queue[1]
	if cq.type ~= "SAY" then return end
	
	-- we don't verify the message contents because it can be affected if the player is drunk
	if guid == UnitGUID( "player" ) then
		self:ChatConfirmed()
	end
end

-------------------------------------------------------------------------------
function Main:OnChatMsgEmote( event, message, sender, _, _, _, _, _, _, _, _, _, guid )

	if #g_chat_queue == 0 then return end
	local cq = g_chat_queue[1]
	if cq.type ~= "EMOTE" then return end
	
	if guid == UnitGUID( "player" ) then 
		self:ChatConfirmed() 
	end
end

-------------------------------------------------------------------------------
function Main:OnChatMsgYell( event, message, sender, _, _, _, _, _, _, _, _, _, guid )

	if #g_chat_queue == 0 then return end
	local cq = g_chat_queue[1]
	if cq.type ~= "YELL" then return end
	
	if guid == UnitGUID( "player" ) then 
		self:ChatConfirmed() 
	end
end

-------------------------------------------------------------------------------
function Main:OnChatMsgSystem( event, message, sender, _, _, target )

	if #g_chat_queue == 0 then 
		-- we aren't expecting anything
		return 
	end
	
	if message == ERR_CHAT_THROTTLED and sender == "" then
		-- we got a throttle error, and we want to retry
		self:ChatFailed()
	end
end
